from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, TypeVar
import abc
import statistics
import enum
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import functools
import dataclasses

import ray.rllib.env.multi_agent_env
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error
import gymnasium
import gymnasium.spaces
from frozendict import frozendict
import numpy as np
from scipy import stats
import more_itertools

from chicken_coop import county
from chicken_coop.county import misc
from chicken_coop.county.typing import Agent, RealNumber
from chicken_coop.county.constants import ALL_AGENTS

from chicken_coop.coop_config import CoopConfig
from .dominance_hierarching import DominanceHierarchy

_T = TypeVar('_T')


@dataclasses.dataclass(frozen=True, kw_only=True)
class BaseArt(abc.ABC):

    @staticmethod
    @abc.abstractmethod
    def sum(items, /) -> BaseArt:
        raise NotImplementedError

    def __add__(self, other: Any) -> BaseArt:
        if not isinstance(other, type(self)):
            return NotImplemented
        return type(self).sum(self, other)

    @classmethod
    @functools.cache
    def _get_summable_fields(cls):
        return {field_name: field for field_name, field in cls.__dataclass_fields__.items()
                if field_name != 'coop_config'}




@dataclasses.dataclass(frozen=True, kw_only=True)
class AgentPortrait(BaseArt):
    output_fields = (
        'aggressiveness', 'reward', 'weight', #'rapport',
        # 'anguish', 'boredom',
    )

    n_victories: int
    n_defeats: int
    n_anguishes: int
    n_boredoms: int

    coop_config: CoopConfig

    @functools.cached_property
    def weight(self) -> int:
        return (self.n_victories + self.n_defeats + self.n_anguishes + self.n_boredoms)

    # This definition might not be equivalent to the rapport definition for a dialog, so
    # it's commented out.
    # @functools.cached_property
    # def rapport(self) -> RealNumber:
        # if self.weight == 0:
            # return 0
        # else:
            # return (self.n_victories + self.n_defeats) / self.weight

    @functools.cached_property
    def anguish(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return self.n_anguishes / self.weight

    @functools.cached_property
    def boredom(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return self.n_boredoms / self.weight

    @functools.cached_property
    def aggressiveness(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return (self.n_victories + self.n_anguishes) / self.weight

    @functools.cached_property
    def reward(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return (
                self.n_victories * self.coop_config.victory_reward +
                self.n_defeats * self.coop_config.defeat_reward +
                self.n_anguishes * self.coop_config.anguish_reward +
                self.n_boredoms * self.coop_config.boredom_reward
            ) / self.weight

    @staticmethod
    def sum(items: Iterable[AgentPortrait], /) -> AgentPortrait:
        agent_portraits = tuple(items)
        (coop_config,) = {agent_portrait.coop_config for agent_portrait in agent_portraits}
        return AgentPortrait(
            **{
                field: sum(getattr(agent_portrait, field) for agent_portrait in agent_portraits)
                for field in AgentPortrait._get_summable_fields()
            },
            coop_config=coop_config,
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Dialog(BaseArt):
    output_fields = (
        #'anguish', 'boredom',
        'aggressiveness', 'reward', 'polarity', 'purity', 'rapport', 'weight',
        'left_aggressiveness', 'right_aggressiveness',
    )

    n_agent_a_victories: int
    n_agent_b_victories: int
    n_anguishes: int
    n_boredoms: int

    coop_config: CoopConfig

    @functools.cached_property
    def n_agent_a_hawks(self) -> int:
        return (self.n_agent_a_victories + self.n_anguishes)

    @functools.cached_property
    def n_agent_b_hawks(self) -> int:
        return (self.n_agent_b_victories + self.n_anguishes)

    @functools.cached_property
    def left_aggressiveness(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return self.n_agent_a_hawks / self.weight

    @functools.cached_property
    def right_aggressiveness(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return self.n_agent_b_hawks / self.weight

    @functools.cached_property
    def weight(self) -> int:
        return (self.n_agent_a_victories + self.n_agent_b_victories + self.n_anguishes +
                self.n_boredoms)

    @functools.cached_property
    def polarity(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return (self.n_agent_b_hawks - self.n_agent_a_hawks) / self.weight

    @functools.cached_property
    def rapport(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return abs(self.n_agent_a_hawks - self.n_agent_b_hawks) / self.weight
            # return (self.n_agent_b_victories + self.n_agent_a_victories) / self.weight

    @functools.cached_property
    def anguish(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return self.n_anguishes / self.weight

    @functools.cached_property
    def boredom(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return self.n_boredoms / self.weight

    @functools.cached_property
    def aggressiveness(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return ((self.n_agent_a_victories + self.n_agent_b_victories + self.n_anguishes) /
                                                                                        self.weight)

    @functools.cached_property
    def purity(self) -> RealNumber:
        return abs(self.polarity)

    @functools.cached_property
    def agent_portraits(self) -> tuple[AgentPortrait, AgentPortrait]:
        return (
            AgentPortrait(
                n_victories=self.n_agent_a_victories,
                n_defeats=self.n_agent_b_victories,
                n_anguishes=self.n_anguishes,
                n_boredoms=self.n_boredoms,
                coop_config=self.coop_config,
            ),
            AgentPortrait(
                n_victories=self.n_agent_b_victories,
                n_defeats=self.n_agent_a_victories,
                n_anguishes=self.n_anguishes,
                n_boredoms=self.n_boredoms,
                coop_config=self.coop_config,
            ),
        )


    @functools.cached_property
    def reward(self) -> RealNumber:
        return statistics.mean(agent_portrait.reward for agent_portrait in self.agent_portraits)


    @staticmethod
    def make_zero(coop_config: CoopConfig) -> Dialog:
        return Dialog(n_agent_a_victories=0, n_agent_b_victories=0, n_anguishes=0, n_boredoms=0,
                      coop_config=coop_config)

    @staticmethod
    def sum(items: Iterable[Dialog], /) -> Dialog:
        '''
        Sum a few dialog objects, adding all the metrics together.

        This should only be used on dialogs between the same two agents, in the same order. This
        isn't enforced so you should be careful to only feed dialogs of the same agent pair to this
        function.
        '''
        dialogs = tuple(items)
        return Dialog(
            **{
                field: sum(getattr(dialog, field) for dialog in dialogs)
                for field in Dialog._get_summable_fields()
            },
            coop_config=dialogs[0].coop_config
        )

    def to_array(self) -> np.ndarray:
        return np.array(
            tuple(getattr(self, field) for field in Dialog._get_summable_fields()),
            dtype=float
        )

    @staticmethod
    def from_array(array: np.ndarray, coop_config: CoopConfig) -> Dialog:
        return Dialog(
            **dict(
                zip(
                    Dialog._get_summable_fields(),
                    map(float, array)
                )
            ),
            coop_config=coop_config,
        )

    @staticmethod
    def _dict_to_array(dialog_by_i_agent_pair: dict[_T, Dialog]) -> np.ndarray:
        (coop_config,) = {dialog.coop_config for dialog in dialog_by_i_agent_pair.values()}
        assert tuple(dialog_by_i_agent_pair) == tuple(coop_config.i_agent_pairs_full)
        return np.array(
            [dialog.to_array() for dialog in dialog_by_i_agent_pair.values()],
            dtype=float
        )

    @staticmethod
    def _array_to_dict(array: np.ndarray, coop_config: CoopConfig) -> dict[_T, Dialog]:
        n_agents = misc.int_sqrt(array.shape[0])
        i_agent_pairs_full = itertools.product(range(n_agents), repeat=2)
        from_array = lambda a_: Dialog.from_array(a_, coop_config=coop_config)
        return dict(zip(i_agent_pairs_full, map(from_array, array)))

    def __neg__(self) -> Dialog:
        return Dialog(
            n_agent_a_victories=self.n_agent_b_victories,
            n_agent_b_victories=self.n_agent_a_victories,
            n_anguishes=self.n_anguishes,
            n_boredoms=self.n_boredoms,
            coop_config=self.coop_config
        )


@dataclasses.dataclass(frozen=True, kw_only=True)
class Zeitgeist(BaseArt):
    dialog_by_i_agent_pair: dict[tuple[int, int], Dialog]

    def __post_init__(self) -> None:
        assert more_itertools.is_sorted(self.dialog_by_i_agent_pair)


    @functools.cached_property
    def coop_config(self) -> CoopConfig:
        (coop_config,) = {dialog.coop_config for dialog in self.dialog_by_i_agent_pair.values()}
        return coop_config

    @functools.cached_property
    def n_agents(self) -> int:
        return self.coop_config.n_agents

    # @staticmethod
    # def make_zero(coop_config: CoopConfig) -> Dialog:
        # return Zeitgeist(
            # dialog_by_i_agent_pair=dict(zip(coop_config.i_agent_pairs_full,
                                            # itertools.repeat(Dialog.make_zero()))),
        # )

    @functools.cached_property
    def dialog_by_rising_i_agent_pair(self) -> dict[tuple[int, int], Dialog]:
        return {(i_agent_a, i_agent_b): dialog for (i_agent_a, i_agent_b), dialog
                in self.dialog_by_i_agent_pair.items() if i_agent_a < i_agent_b}

    @functools.cached_property
    def weight(self) -> int:
        return sum(dialog.weight for (i_agent_a, i_agent_b), dialog in
                   self.dialog_by_rising_i_agent_pair.items())

    @functools.cached_property
    def agent_portraits(self) -> tuple[AgentPortrait, ...]:
        addend_agent_portraits_by_i_agent = [[] for _ in range(self.n_agents)]
        for (i_agent, _), dialog in self.dialog_by_i_agent_pair.items():
            addend_agent_portraits_by_i_agent[i_agent].append(dialog.agent_portraits[0])
        return tuple(
            map(
                AgentPortrait.sum,
                addend_agent_portraits_by_i_agent,
            )
        )

    @functools.cached_property
    def population_portrait(self) -> PopulationPortrait:
        return PopulationPortrait(self)

    @staticmethod
    def sum(items: Iterable[Zeitgeist], /) -> Zeitgeist:
        zeitgeists = tuple(items)
        (coop_config,) = {zeitgeist.coop_config for zeitgeist in zeitgeists}
        return Zeitgeist.make_from_dialog_by_i_agent_pair(
            {i_agent_pair: Dialog.sum(zeitgeist.dialog_by_i_agent_pair[i_agent_pair]
                                       for zeitgeist in zeitgeists)
            for i_agent_pair in coop_config.i_agent_pairs_full}
        )

    @staticmethod
    def make_from_dialog_by_i_agent_pair(dialog_by_i_agent_pair: dict[tuple[int, int], Dialog]
                                         ) -> Zeitgeist:
        dialog_by_i_agent_pair = dict(sorted(dialog_by_i_agent_pair.items()))
        return Zeitgeist(dialog_by_i_agent_pair=dialog_by_i_agent_pair)


    def to_array(self) -> np.ndarray:
        assert tuple(self.dialog_by_i_agent_pair) == tuple(self.coop_config.i_agent_pairs_full)
        return np.array(
            [dialog.to_array() for dialog in self.dialog_by_i_agent_pair.values()],
            dtype=float
        )

    @staticmethod
    def from_array(array: np.ndarray, coop_config: CoopConfig) -> Dialog:
        # todo: This array could be shortened by using something with less redundancies than
        # `i_agent_pairs_full`
        n_agents = misc.int_sqrt(array.shape[0])
        i_agent_pairs_full = itertools.product(range(n_agents), repeat=2)
        from_array = lambda a_: Dialog.from_array(a_, coop_config=coop_config)
        return Zeitgeist.make_from_dialog_by_i_agent_pair(
            dict(zip(i_agent_pairs_full, map(from_array, array), strict=True))
        )

    @functools.cached_property
    def output(self) -> dict[str, RealNumber]:
        return {
            **{
                f'population.{output_field}': getattr(self.population_portrait, output_field)
                for output_field in PopulationPortrait.output_fields
            },
            **{
                f'dialog.{i_agent_a}.{i_agent_b}.{field}':
                    getattr(self.dialog_by_i_agent_pair[i_agent_a, i_agent_b], field)
                for i_agent_a, i_agent_b in self.coop_config.i_agent_pairs_without_diagonal
                for field in Dialog.output_fields
            },
            **{
                f'agent.{i_agent}.{field}': getattr(self.agent_portraits[i_agent], field)
                for i_agent in range(self.n_agents)
                for field in AgentPortrait.output_fields
            },
        }

@dataclasses.dataclass(frozen=True)
class PopulationPortrait:
    output_fields = (
        # 'anguish', 'boredom',
        'aggressiveness', 'reward', 'polarity', 'purity', 'pairwise_purity', 'transitivity',
        'rapport', 'weight', 'dominance_hierarchy',
        'restricted_distance_from_visitor_dominance_hierarchy',
    )

    zeitgeist: Zeitgeist

    @functools.cached_property
    def coop_config(self) -> CoopConfig:
        return self.zeitgeist.coop_config

    @functools.cached_property
    def n_agents(self) -> int:
        return self.zeitgeist.n_agents

    @functools.cached_property
    def weight(self) -> int:
        return self.zeitgeist.weight

    @functools.cached_property
    def polarity(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return np.average(
                tuple(dialog.polarity for dialog in
                      self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
                weights=tuple(dialog.weight for dialog in
                              self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
            )


    @functools.cached_property
    def rapport(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return np.average(
                tuple(dialog.rapport for dialog in
                      self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
                weights=tuple(dialog.weight for dialog in
                              self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
            )

    @functools.cached_property
    def anguish(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return np.average(
                tuple(dialog.anguish for dialog in
                      self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
                weights=tuple(dialog.weight for dialog in
                              self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
            )

    @functools.cached_property
    def boredom(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return np.average(
                tuple(dialog.boredom for dialog in
                      self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
                weights=tuple(dialog.weight for dialog in
                              self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
            )

    @functools.cached_property
    def aggressiveness(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return np.average(
                tuple(dialog.aggressiveness for dialog in
                      self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
                weights=tuple(dialog.weight for dialog in
                              self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
            )


    @functools.cached_property
    def purity(self) -> RealNumber:
        return abs(self.polarity)


    @functools.cached_property
    def pairwise_purity(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            return np.average(
                tuple(dialog.purity for dialog in
                      self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
                weights=tuple(dialog.weight for dialog in
                              self.zeitgeist.dialog_by_rising_i_agent_pair.values()),
            )

    @functools.cached_property
    def transitivity(self) -> RealNumber:
        if self.weight == 0:
            return 0
        else:
            agent_gifts = [
                abs(
                    sum(self.zeitgeist.dialog_by_i_agent_pair[j_agent, i_agent].polarity
                        for j_agent in set(range(self.n_agents)) - {i_agent})
                ) for i_agent in range(self.n_agents)
            ]
            max_agent_gifts = tuple(map(abs,
                                        range(self.n_agents - 1, -self.n_agents, -2)))
            assert len(agent_gifts) == len(max_agent_gifts)
            assert max_agent_gifts[0] == max_agent_gifts[-1] == self.n_agents - 1
            transitivity = sum(agent_gifts) / sum(max_agent_gifts)
            assert 0 <= transitivity <= 1.000001, (f"Transitivity must be between 0 and 1 but it's "
                                                   f"{transitivity}")
            return transitivity

    @functools.cached_property
    def reward(self):
        return statistics.mean(agent_portrait.reward for agent_portrait in
                               self.zeitgeist.agent_portraits)

    @functools.cached_property
    def dominance_hierarchy(self) -> DominanceHierarchy:

        return DominanceHierarchy(
            (
                i_agent_pair for (i_agent_pair,
                                  dialog) in self.zeitgeist.dialog_by_i_agent_pair.items()
                if dialog.polarity <= - self.coop_config.dominance_hierarchy_polarity_threshold
            ),
            n_agents=self.n_agents
        )

    @functools.cached_property
    def restricted_distance_from_visitor_dominance_hierarchy(self) -> Optional[RealNumber]:
        visitor_dominance_hierarchy = self.coop_config.visitor_dominance_hierarchy
        if visitor_dominance_hierarchy is None:
            return None
        else:
            return self.dominance_hierarchy.get_distance(
                visitor_dominance_hierarchy, i_agents=self.coop_config.i_resident_agents
            )






from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, Sequence
import itertools
import pathlib
import statistics
import pickle
import base64
import time
import abc
import lzma
import logging
import functools
import collections
import sys
import operator
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string

import more_itertools
import ray.rllib.env.multi_agent_env
import ray.rllib.algorithms.callbacks
from ray.rllib.utils.typing import PolicyID, MultiAgentDict, ModelWeights
from ray.rllib import algorithms
import ray.rllib.algorithms.a2c
import ray.rllib.algorithms.a3c
import ray.rllib.algorithms.ppo
import ray.rllib.algorithms.sac
import ray.rllib.algorithms.appo
import ray.rllib.algorithms.impala
import ray.rllib.algorithms.maddpg
from ray import rllib
import gymnasium
import gymnasium.spaces
import numpy as np
import pysnooper

from chicken_coop import county
from chicken_coop.county import misc
from chicken_coop.county import misc
from chicken_coop.county import csv_tools
from chicken_coop.county import policing
import chicken_coop
from chicken_coop.county.typing import Agent, RealNumber, PolicyID
from .stating import State, Move, Dialog
from .enving import Env


class Callbackino(ray.rllib.algorithms.callbacks.DefaultCallbacks):
    def on_episode_end(
        self,
        *,
        worker: ray.rllib.evaluation.RolloutWorker,
        base_env: ray.rllib.BaseEnv,
        policies: dict[PolicyID, ray.rllib.Policy],
        episode: ray.rllib.evaluation.Episode | ray.rllib.evaluation.episode_v2.EpisodeV2 |
            Exception,
        **kwargs,
        ) -> None:

        if isinstance(episode, Exception):
            return


    def on_train_result(
        self,
        *,
        algorithm: ray.rllib.algorithms.Algorithm,
        result: dict,
        **kwargs,
        ) -> None:
        """Called at the end of Algorithm.train().

        Args:
            algorithm: Current Algorithm instance.
            result: Dict of results returned from Algorithm.train() call.
                You can mutate this object to add additional metrics.
            kwargs: Forward compatibility placeholder.
        """
        pass

    def on_episode_end(
        self,
        *,
        worker: rllib.evaluation.RolloutWorker,
        base_env: rllib.env.BaseEnv,
        policies: dict[PolicyID, rllib.Policy],
        episode: Union[rllib.evaluation.Episode, rllib.evaluation.episode_v2.EpisodeV2, Exception],
        env_index: Optional[int] = None,
        **kwargs,
    ) -> None:
        """Runs when an episode is done.

        Args:
            worker: Reference to the current rollout worker.
            base_env: BaseEnv running the episode. The underlying
                sub environment objects can be retrieved by calling
                `base_env.get_sub_environments()`.
            policies: Mapping of policy id to policy
                objects. In single agent mode there will only be a single
                "default_policy".
            episode: Episode object which contains episode
                state. You can use the `episode.user_data` dict to store
                temporary data, and `episode.custom_metrics` to store custom
                metrics for the episode.
                In case of environment failures, episode may also be an Exception
                that gets thrown from the environment before the episode finishes.
                Users of this callback may then handle these error cases properly
                with their custom logics.
            env_index: The index of the sub-environment that ended the episode
                (within the vector of sub-environments of the BaseEnv).
            kwargs: Forward compatibility placeholder.
        """
        (env,) = base_env.get_sub_environments()
        last_state: State = env.state
        episode.custom_metrics |= {
            'zeitgeist_array': last_state.zeitgeist.to_array()
            # f'dialog.{agent_a}.{agent_b}': last_state.dialog_by_i_agent_pair[i_agent_a, i_agent_b]
            # for (i_agent_a, agent_a), (i_agent_b, agent_b) in
            # itertools.combinations(enumerate(self.coop_config.policy_by_agent), 2)
        }


from __future__ import annotations

import csv
import pathlib
import contextlib
from typing import Mapping, Any, Optional, ContextManager, Iterable, Type
from types import TracebackType as Traceback


class CSVWriter(ContextManager):
    def __init__(self, path: str | pathlib.Path, fields: Iterable[str], *,
                 overwrite: bool = True) -> None:
        self.path = pathlib.Path(path)
        self.fields = tuple(fields)
        self.overwrite = overwrite

    def __enter__(self) -> CSVWriter:
        self._file = self.path.open('w' if self.overwrite else 'a', newline='')
        self._dict_writer = csv.DictWriter(self._file, fieldnames=self.fields)
        if self.overwrite:
            self._dict_writer.writeheader()
            self._file.flush()
        return self

    def __exit__(self, exc_type: Optional[Type[BaseException]], exc_value: Optional[BaseException],
                 traceback: Optional[Traceback]) -> Optional[bool]:
        try:
            self._dict_writer = None
        finally:
            self._file.close()
            self._file = None

    def write_row(self, row: Mapping[str, Any], **row_extra: Any) -> None:
        self._dict_writer.writerow(dict(row) | row_extra)
        self._file.flush()

    def write_rows(self, rows: Iterable[Mapping[str, Any]]) -> None:
        self._dict_writer.writerows(rows)
        self._file.flush()




from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, Container
import abc
import io
import pickle
import lzma
import time
import pathlib
import collections
import dataclasses
import hashlib
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import copy

import numpy as np
import ray.rllib.env.multi_agent_env
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error

from chicken_coop.county import misc


def normalize_weights(weights: dict[str, np.ndarray]) -> dict[str, np.ndarray]:
    first_key = sorted(weights)[0]
    if (worker_snippet_match := re.search('_wk[0-9]+(?=/)', first_key)):
        fixed_weights = {}
        worker_snippet = worker_snippet_match.group(0)
        for key, value in weights.items():
            assert key.count(worker_snippet) == 1
            fixed_key = key.replace(worker_snippet, '')
            assert len(key) - len(fixed_key) == len(worker_snippet)
            fixed_weights[fixed_key] = value.copy()
        return fixed_weights
    else:
        return copy.deepcopy(weights)



class PolicySnapshot:
    def __init__(self, policy_or_weights: Mapping[str, np.ndarray] | ray.rllib.Policy) -> None:
        self.weights = normalize_weights(
            policy_or_weights.get_weights()
            if isinstance(policy_or_weights, ray.rllib.Policy)
            else policy_or_weights
        )
        self.hash = hashlib.sha512(
            b''.join(
                array.tobytes() for array in tuple(zip(*sorted(self.weights.items())))[1]
            )
        ).hexdigest()[:6]


    def __eq__(self, other: Any) -> bool:
        if isinstance(other, PolicySnapshot):
            return (
                (self.weights.keys() == other.weights.keys()) and
                all(np.array_equal(value, other.weights[key])
                    for key, value in self.weights.items())
            )
        else:
            return NotImplemented

    def __hash__(self) -> str:
        return self.hash

    def __repr__(self) -> str:
        return f'<{type(self).__name__}: {self.hash}>'


    def transplant_to_algorithm(self,
                                algorithm: ray.rllib.algorithms.Algorithm,
                                policy_name: str
                                ) -> None:
        CultureSnapshot({policy_name: self,}).transplant_to_algorithm(algorithm)

    def compickle_to_bytes(self) -> bytes:
        return misc.compickle_to_bytes(self.weights)


    @staticmethod
    def uncompickle_from_bytes(compickled_bytes: bytes) -> PolicySnapshot:
        return PolicySnapshot(misc.uncompickle_from_bytes(compickled_bytes))

    def __neg__(self) -> PolicySnapshot:
        return PolicySnapshot({key: -value for key, value in self.weights.items()})



class CultureSnapshot(Mapping):
    def __init__(self, policy_snapshot_by_policy_name: Mapping[str, PolicySnapshot]) -> None:
        self.policy_snapshot_by_policy_name = dict(policy_snapshot_by_policy_name)

    __getitem__ = lambda self, key: self.policy_snapshot_by_policy_name[key]
    __len__ = lambda self: len(self.policy_snapshot_by_policy_name)
    __iter__ = lambda self: iter(self.policy_snapshot_by_policy_name)
    __repr__ = lambda self: f'{type(self).__name__}({self.policy_snapshot_by_policy_name})'


    @staticmethod
    def import_from_algorithm(algorithm: ray.rllib.algorithms.Algorithm, *,
                              exclude_policy_names: Container[str] = ()) -> CultureSnapshot:
        return CultureSnapshot(
            {
                policy_name: PolicySnapshot(policy)
                for policy_name, policy in
                algorithm.workers.local_worker().policy_map.items()
                if policy_name not in exclude_policy_names
            }
        )


    def transplant_to_algorithm(self,
                            algorithm: ray.rllib.algorithms.Algorithm,
                            ) -> None:
        for policy_name, policy_snapshot in self.items():
            algorithm.get_policy(policy_name).set_weights(policy_snapshot.weights)
        algorithm.workers.sync_weights(self)
        workers = algorithm.workers.remote_workers()
        for worker in workers:
            ray.get(
                worker.set_weights.remote({
                    policy_name: policy_snapshot.weights for
                    policy_name, policy_snapshot in self.items()
                })
            )
        for i in range(10):
            time.sleep(i)
            for worker, (policy_name, policy_snapshot) in itertools.product(workers,
                                                                            self.items()):
                policy_weights_on_worker = ray.get(
                                                worker.get_weights.remote(policy_name))[policy_name]
                if PolicySnapshot(policy_weights_on_worker) != policy_snapshot:
                    break # Break out of the small loop to restart the big loop and sleep.
            else:
                # Loop wasn't broken out of, so we confirmed that all policies were propogated to
                # remote workers.
                return

        raise RuntimeError("Policies didn't propogate to all remote workers.")

    def save(self, file_or_path: pathlib.Path | io.BufferedWriter) -> None:
        misc.compickle(self, file_or_path)

    @staticmethod
    def load(file_or_path: pathlib.Path | io.BufferedWriter) -> CultureSnapshot:
        result = misc.uncompickle(file_or_path)
        if not isinstance(result, CultureSnapshot):
            raise ValueError
        return result

    def __neg__(self) -> CultureSnapshot:
        return CultureSnapshot(
            {policy_name: - policy_snapshot for policy_name, policy_snapshot in self.items()}
        )



import logging

import json_fix # This activates `json_fix`.

from . import filtros
filtros.activate()

from .base_env import BaseEnv
from .base_state import BaseState
from . import constants

def init_ray():
    import ray
    if not ray._private.worker.global_worker.connected:
        ray.init(logging_level=logging.WARN)



# pytype: disable=module-attr
from __future__ import annotations

import collections
from typing import Tuple, Optional, Dict, Mapping, Iterable, Callable, Any
import itertools
import abc
import ray.rllib.env.multi_agent_env

from .base_state import BaseState
from chicken_coop.county.constants import ALL_AGENTS
from chicken_coop.county.typing import Agent, Action, Observation
from chicken_coop.county.misc import compute_actions_for_all_agents


class BaseEnv(ray.rllib.env.multi_agent_env.MultiAgentEnv, abc.ABC):

    def __init__(self, config: Optional[Mapping] = None) -> None:
        ray.rllib.env.multi_agent_env.MultiAgentEnv.__init__(self)
        self.config = config = (config or {})
        self.states = []

    @abc.abstractmethod
    def make_initial_state(self) -> BaseState:
        raise NotImplementedError

    def reset(self, *, seed: Optional[int] = None,
              options: Optional[Mapping] = None) -> Tuple[Mapping[Agent, Observation],
                                                          Mapping[Agent, Any]]:
        match (initial_states_raw := (options or {}).get('initial_states', None)):
            case BaseState():
                self.states[:] = (initial_states_raw,)
            case collections.abc.Iterable():
                self.states[:] = initial_states_raw
            case None:
                self.states[:] = [self.make_initial_state()]
            case _:
                raise RuntimeError

        return (self.state.observation_by_agent, collections.defaultdict(lambda: None))

    @property
    def state(self) -> BaseState:
        return self.states[-1]

    @property
    def observation_by_agent(self) -> Mapping[Agent, Observation]:
        return self.states[-1].observation_by_agent

    def step(self, actions: Mapping[Agent, Action]) -> tuple[Mapping, Mapping, Mapping, Mapping]:
        self.states.append(self.state.step(actions))
        return (self.state.observation_by_agent, self.state.reward_by_agent,
                self.state.terminated_by_agent, self.state.truncated_by_agent, {})

    def render(self, mode: Optional[str] = None) -> str:
        return self.state.text

    def play(self,
             algorithm: ray.rllib.algorithms.Algorithm,
             n: Optional[int] = None,
             stop_condition: Optional[Callable[BaseState, bool]] = None,
             ) -> Iterator[BaseState]:
        for i in (range(n) if n else itertools.count()):
            if stop_condition is not None and stop_condition(self.state):
                return
            actions = compute_actions_for_all_agents(algorithm, self)
            _, _, terminated_by_agent, truncated_by_agent, _ = self.step(actions)
            yield self.state
            if terminated_by_agent[ALL_AGENTS] or truncated_by_agent[ALL_AGENTS]:
                return

    @classmethod
    def sample_episode(cls,
                       algorithm: ray.rllib.algorithms.Algorithm,
                       n: Optional[int] = None,
                       stop_condition: Optional[Callable[BaseState, bool]] = None,
                       initial_states: None | BaseState | Iterable[BaseState] = None,
                       ) -> Iterator[BaseState]:
        env = cls(config=algorithm.config['env_config'])
        env.reset(options={'initial_states': initial_states})
        yield env.state
        new_n = None if (n is None) else (n - 1)
        yield from env.play(algorithm, n=new_n, stop_condition=stop_condition)





from __future__ import annotations

from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, TypeVar, Annotated
import collections
import pathlib
import abc
import math
import statistics
import enum
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import functools
import dataclasses

import ray.rllib.env.multi_agent_env
from ray.rllib.algorithms.ppo import PPO # pytype: disable=import-error
import gymnasium
import gymnasium.spaces
from frozendict import frozendict
import numpy as np
from scipy import stats
import more_itertools
import networkx as nx
import plotly.graph_objects as go

from chicken_coop import county
from chicken_coop.county import misc
from chicken_coop.county import nx_tools
from chicken_coop.county.typing import Agent, RealNumber
from chicken_coop.county.constants import ALL_AGENTS

from .exceptions import CoopException


class DominanceHierarchyException(CoopException):
    pass

class InvalidDominanceHierarchy(DominanceHierarchyException):
    pass

class AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy(DominanceHierarchyException):
    pass

class AttemptedCompleteOperationOnIncompleteDominanceHierarchy(DominanceHierarchyException):
    pass

class DominanceRole(enum.Enum):
    DOMINANT = 'dominant'
    SUBORDINATE = 'subordinate'


class DominanceHierarchy:

    _initialized = False

    def __init__(self, edges_or_concise_text: Iterable[tuple[int, int]] | str = (), /, *,
                 n_agents: Optional[int] = None) -> None:
        self.di_graph = nx.DiGraph()
        if isinstance(edges_or_concise_text, str):
            self.di_graph = nx_tools.concise_text_to_oriented_graph(edges_or_concise_text)
        else:
            self.di_graph = nx.DiGraph(edges_or_concise_text)
            if n_agents is not None:
                self.di_graph.add_nodes_from(tuple(i for i in range(n_agents)
                                                   if i not in self.di_graph))
            else:
                self.isolated_i_agents = ()
        self.isolated_i_agents = tuple(nx_tools.nx.isolates(self.di_graph))
        if n_agents is not None:
            assert set(self.di_graph) == set(range(n_agents))

        self.condensed_di_graph = nx.condensation(self.di_graph)
        self.i_agents = tuple(sorted(self.di_graph.nodes))
        assert len(self.i_agents) == self.i_agents[-1] + 1
        self._rising_agent_pairs = tuple(itertools.combinations(self.i_agents, 2))
        self.n_agents = len(self.i_agents)
        if n_agents is not None:
            assert self.n_agents == n_agents
        self.edges = tuple(sorted(self.di_graph.edges))
        for (i_agent, j_agent) in self._rising_agent_pairs:
            assert not (self.di_graph.has_edge(i_agent, j_agent) and
                        self.di_graph.has_edge(j_agent, i_agent))
        self._initialized = True

    def __len__(self) -> int:
        return self.n_agents

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, DominanceHierarchy):
            return False
        return (type(self) is type(other)) and nx.utils.graphs_equal(self.di_graph, other.di_graph)

    def __hash__(self) -> int:
        return hash((type(self), str(self)))

    def __or__(self, other: Any) -> DominanceHierarchy:
        if not isinstance(other, DominanceHierarchy):
            return NotImplemented
        return DominanceHierarchy(set(self.edges) | set(other.edges))

    def __sub__(self, other: Any) -> int:
        if not isinstance(other, DominanceHierarchy):
            return NotImplemented
        return self.get_distance(other)


    def __str__(self) -> str:
        return nx_tools.oriented_graph_to_concise_text(self.di_graph)

    def __repr__(self) -> str:
        if self._initialized:
            return f'{type(self).__name__}({repr(str(self))})'
        else:
            return f'<{type(self).__name__} (uninitialized)>'

    # def __json__(self) -> list[Annotated[list[int], 2]]:
        # return list(map(list, self.edges))

    def __json__(self) -> str:
        return str(self)

    def get_distance(self, other_dominance_hierarchy: DominanceHierarchy, /, *,
                     i_agents: Optional[tuple[int, ...]] = None) -> int:
        if self.n_agents != other_dominance_hierarchy.n_agents:
            raise NotImplementedError

        if i_agents is None:
            i_agents = self.i_agents
        else:
            assert set(i_agents) <= set(self.i_agents)

        assert len(i_agents) >= 2

        rising_agent_pairs = tuple(itertools.combinations(i_agents, 2))

        result = sum((self.get_relationship(i_agent, j_agent) !=
                                       other_dominance_hierarchy.get_relationship(i_agent, j_agent))
                     for (i_agent, j_agent) in rising_agent_pairs) / len(rising_agent_pairs)
        assert 0 <= result <= 1
        return result

    @functools.cached_property
    def simple_cycles(self) -> tuple[tuple[int, ...], ...]:
        return tuple(map(tuple, nx.simple_cycles(self.di_graph)))

    @functools.cached_property
    def is_transitive(self) -> bool:
        return not self.simple_cycles

    @functools.cached_property
    def is_complete(self) -> bool:
        assert self._initialized
        return nx_tools.is_tournament(self.di_graph)

    @functools.cached_property
    def i_agent_by_rank(self) -> tuple[int, ...]:
        if not self.is_transitive:
            raise AttemptedTransitiveOperationOnIntransitiveDominanceHierarchy
        return tuple(nx.topological_sort(self.di_graph))

    @functools.cached_property
    def linear_ranks(self) -> tuple[int, ...]:
        linear_ranks = []
        i_agent_needle = 0
        for condensed_node in nx.topological_sort(self.condensed_di_graph):
            nodes = self.condensed_di_graph.nodes.data()[condensed_node]['members']
            if len(nodes) == 1:
                linear_ranks.append(i_agent_needle)
            i_agent_needle += len(nodes)

        assert i_agent_needle == self.n_agents
        return tuple(linear_ranks)

    @functools.cached_property
    def non_linear_ranks(self) -> tuple[int, ...]:
        return tuple(i for i in self.i_agents if i not in self.linear_ranks)

    @functools.cached_property
    def condensed_plot_svg(self) -> go.Figure:
        from .plotting.plotting_dominance_hierarchies import make_condensed_plot_svg
        return make_condensed_plot_svg(self)

    @staticmethod
    def get_from_rollout_jsonl(rollout_jsonl_path: pathlib.Path) -> DominanceHierarchy:
        return DominanceHierarchy(
            misc.get_last_entry_from_jsonl(rollout_jsonl_path)['population']['dominance_hierarchy']
        )

    def get_relationship(self, i_agent: int, i_other_agent: int) -> Optional[DominanceRole]:
        assert i_agent != i_other_agent
        assert {i_agent, i_other_agent} <= set(self.i_agents)
        if self.di_graph.has_edge(i_agent, i_other_agent):
            return DominanceRole.DOMINANT
        elif self.di_graph.has_edge(i_other_agent, i_agent):
            return DominanceRole.SUBORDINATE
        return None

    @staticmethod
    def get_trivial(n_agents: int) -> DominanceHierarchy:
        return DominanceHierarchy(' => '.join(map(str, range(n_agents))))


    @staticmethod
    def get_mean_rank_linearity(
        dominance_hierarchies: Iterable[DominanceHierarchy] | DominanceHierarchy
        ) -> tuple[RealNumber, ...]:
        dominance_hierarchies = misc.to_tuple(dominance_hierarchies, item_type=DominanceHierarchy)
        (n_ranks,) = set(map(len, dominance_hierarchies))
        linear_ranks = tuple(
            itertools.chain.from_iterable(dominance_hierarchy.linear_ranks for
                                          dominance_hierarchy in dominance_hierarchies)
        )
        counter = collections.Counter(linear_ranks)
        return tuple(counter[i] / len(dominance_hierarchies) for i in range(n_ranks))










from __future__ import annotations

from typing import (Tuple, Optional, Dict, Mapping, Iterable, Any, Sequence, Callable, Hashable,
                    Callable)
import typing
import pathlib
import copy
import enum
import yaml
import datetime as datetime_module
import statistics
import pickle
import base64
import time
import abc
import lzma
import logging
import functools
import collections
import sys
import operator
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import contextlib

import more_itertools
import ray.rllib.env.multi_agent_env
import ray.rllib.algorithms.callbacks
from ray.rllib.utils.typing import PolicyID, MultiAgentDict, ModelWeights
from ray.rllib import algorithms
import ray.tune
import ray.rllib.algorithms.ppo
import gymnasium
import gymnasium.spaces
import numpy as np
import pysnooper
import click

from chicken_coop import county
from chicken_coop.county.thready_krueger import ThreadyKrueger
from chicken_coop.county import misc
from chicken_coop.county import csv_tools
from chicken_coop.county import policing
import chicken_coop
from chicken_coop.county.typing import Agent, RealNumber
from . import defaults
from .dominance_hierarching import DominanceHierarchy


@dataclasses.dataclass(kw_only=True, repr=False)
class CoopConfig:

    victory_reward: RealNumber = defaults.DEFAULT_VICTORY_REWARD
    boredom_reward: RealNumber = defaults.DEFAULT_BOREDOM_REWARD
    defeat_reward: RealNumber = defaults.DEFAULT_DEFEAT_REWARD
    anguish_reward: RealNumber = defaults.DEFAULT_ANGUISH_REWARD

    episode_length: RealNumber = defaults.DEFAULT_EPISODE_LENGTH
    identity_is_linear: bool = defaults.DEFAULT_IDENTITY_IS_LINEAR
    n_agents: int = defaults.DEFAULT_N_AGENTS
    n_generations: int = defaults.DEFAULT_N_GENERATIONS
    train_batch_size: int = defaults.DEFAULT_TRAIN_BATCH_SIZE
    learning_rate: RealNumber = defaults.DEFAULT_LEARNING_RATE
    n_rollout_workers: int = defaults.DEFAULT_N_ROLLOUT_WORKERS
    policy_snapshot_period: int = defaults.DEFAULT_POLICY_SNAPSHOT_PERIOD
    flip_initial_weights: bool = defaults.DEFAULT_FLIP_INITIAL_WEIGHTS
    dominance_hierarchy_polarity_threshold: float = \
                                          defaults.DEFAULT_DOMINANCE_RELATIONSHIP_POLARITY_THRESHOLD
    observation_accuracy: float = defaults.DEFAULT_OBSERVATION_ACCURACY
    i_visitor_agents: tuple[int, ...] = ()
    visitor_dominance_hierarchy: Optional[DominanceHierarchy] = None
    freeze_visitors: bool = defaults.DEFAULT_FREEZE_VISITORS


    def __post_init__(self) -> None:
        self.i_agents = tuple(range(self.n_agents))
        self.i_agent_pairs_full = tuple(itertools.product(self.i_agents, repeat=2))
        self.i_agent_pairs_increasing = tuple(itertools.combinations(self.i_agents, 2))
        self.i_agent_pairs_decreasing = tuple(
            sorted(map(tuple, (map(reversed, itertools.combinations(self.i_agents, 2)))))
        )
        self.i_agent_pairs_diagonal = tuple(zip(self.i_agents, self.i_agents, strict=True))
        self.i_agent_pairs_without_diagonal = tuple(itertools.permutations(self.i_agents, 2))
        self.i_resident_agents = tuple(i_agent for i_agent in self.i_agents
                                       if i_agent not in self.i_visitor_agents)

        self.pad_i_agent = lambda i_agent: str(i_agent).rjust(len(str(self.n_agents)), '0')
        self.policy_by_agent = {f'agent_{self.pad_i_agent(i)}': f'policy_{self.pad_i_agent(i)}'
                                for i in self.i_agents}


        ### Validating: ############################################################################
        #                                                                                          #
        assert len({self.victory_reward, self.boredom_reward, self.defeat_reward,
                    self.anguish_reward}) == 4
        assert self.n_agents % 2 == 0
        assert set(self.i_visitor_agents) <= set(self.i_agents)
        assert more_itertools.is_sorted(self.i_visitor_agents)

        assert all(
            map(more_itertools.is_sorted,
                (self.i_agent_pairs_full, self.i_agent_pairs_increasing,
                 self.i_agent_pairs_decreasing, self.i_agent_pairs_diagonal,
                 self.i_agent_pairs_without_diagonal)
            )
        )
        assert set(self.i_agent_pairs_full) == (
            set(self.i_agent_pairs_increasing) | set(self.i_agent_pairs_decreasing) |
            set(self.i_agent_pairs_diagonal)
        ) == (
            set(self.i_agent_pairs_diagonal) | set(self.i_agent_pairs_without_diagonal)
        )
        assert not (set(self.i_agent_pairs_diagonal) & set(self.i_agent_pairs_without_diagonal))
        assert set(self.i_agent_pairs_without_diagonal) == (set(self.i_agent_pairs_increasing) |
                                                            set(self.i_agent_pairs_decreasing))
        assert not (set(self.i_agent_pairs_increasing) & set(self.i_agent_pairs_decreasing))
        assert not (set(self.i_agent_pairs_increasing) & set(self.i_agent_pairs_diagonal))
        assert not (set(self.i_agent_pairs_diagonal) & set(self.i_agent_pairs_decreasing))
        #                                                                                          #
        ### Finished validating. ###################################################################

    def policy_mapping_fn(self,
                          agent_id: Agent,
                          episode: Optional[ray.rllib.evaluation.episode.Episode] = None,
                          worker: Optional[ray.rllib.evaluation.rollout_worker.
                                           RolloutWorker] = None,
                          **kwargs) -> PolicyID:
        return self.policy_by_agent[agent_id]

    def get_nice_dict(self, *, for_yaml: bool = False) -> dict[str, Any]:
        keys = tuple(key for key in vars(self) if not key.startswith('_') and
                     not key.startswith('i_agent_pairs') and key != 'pad_i_agent') + \
                                                                               ('i_visitor_agents',)
        result = {key: getattr(self, key) for key in keys}
        if for_yaml:
            if self.visitor_dominance_hierarchy is not None:
                result['visitor_dominance_hierarchy'] = str(self.visitor_dominance_hierarchy)
        return result


    @classmethod
    def _get_field_type_for_click(cls, field_name: str) -> type:
        field_type_map = {
            RealNumber: float,
        }
        raw_field_type = typing.get_type_hints(CoopConfig)[field_name]
        try:
            return field_type_map[raw_field_type]
        except KeyError:
            return raw_field_type


    @classmethod
    def add_options_to_click_command(cls,
                                     defaults: Optional[Mapping] = None
                                     ) -> Callable[click.decorators.FC, click.decorators.FC]:
        def inner(command: click.decorators.FC) -> click.decorators.FC:
            defaults_copy = dict(defaults or {})
            if defaults_copy:
                for value in defaults_copy.values():
                    assert isinstance(value, tuple)
            for field in cls.__dataclass_fields__.values():
                field: dataclasses.Field
                field_type = cls._get_field_type_for_click(field.name)
                dashed_field_name = field.name.replace('_', '-')
                option_kwargs = {
                    'type': field_type,
                    'default': defaults_copy.pop(field.name, (field.default,)),
                    'multiple': True,
                    'show_default': True,
                }
                if field_type is bool:
                    option_kwargs['is_flag'] = True
                    name = f'--{dashed_field_name}/--not-{dashed_field_name}'
                else:
                    name = f'--{dashed_field_name}'
                command = click.option(name, **option_kwargs)(command)
            if defaults_copy:
                raise Exception(f'Some unrecognized defaults were passed in: {defaults_copy}')
            return command
        return inner


    def __hash__(self) -> int:
        return hash(
            (type(self),
             *(getattr(self, field) for field in self.__dataclass_fields__))
        )


from __future__ import annotations

import math
import json
import pathlib
import random
import re
import copy
from typing import Iterable

import numpy as np

import ray.rllib.env.multi_agent_env

from chicken_coop.county.typing import Agent, RealNumber
from .stating import State


def triangle_lengths_to_sides(a_length: RealNumber, b_length: RealNumber, c_length: RealNumber
                              ) -> tuple[RealNumber, RealNumber, RealNumber]:
    return (
        math.acos((b_length ** 2 + c_length ** 2 - a_length ** 2) /
                  (2 * b_length * c_length)),
        math.acos((a_length ** 2 + c_length ** 2 - b_length ** 2) /
                  (2 * a_length * c_length)),
        math.acos((a_length ** 2 + b_length ** 2 - c_length ** 2) /
                  (2 * a_length * b_length))
    )

def are_weights_equal(local_weights: dict[str, np.ndarray],
                      worker_weights: dict[str, np.ndarray]) -> bool:
    fixed_worker_weights = copy.deepcopy(worker_weights)
    first_worker_key = sorted(fixed_worker_weights)[0]
    worker_snippet = re.search('_wk[0-9]+(?=/)', first_worker_key).group(0)
    for key, value in tuple(fixed_worker_weights.items()):
        fixed_key = key.replace(worker_snippet, '')
        assert len(key) - len(fixed_key)  == len(worker_snippet)
        fixed_worker_weights[fixed_key] = value
        del fixed_worker_weights[key]

    return (
        set(local_weights) == set(fixed_worker_weights) and
        all(np.array_equal(value, fixed_worker_weights[key])
            for key, value in local_weights.items())
    )


def clamp(number: RealNumber, low: RealNumber, high: RealNumber) -> RealNumber:
    if number < low:
        return low
    elif number > high:
        return high
    else:
        return number


def shuffled(iterable: Iterable[Any]) -> tuple[Any, ...]:
    list_ = list(iterable)
    random.shuffle(list_)
    return tuple(list_)


def get_move_by_state(algorithm: ray.rllib.algorithms.Algorithm,
                      states: Iterable[State],
                      agent: Agent) -> dict[State, int]:
    from .stating import Move
    states = tuple(states)
    policy = algorithm.config['policy_mapping_fn'](agent)
    robot_policy = algorithm.get_policy(policy)
    observation_preprocessor = algorithm.workers.local_worker().preprocessors[policy]
    state_to_flat_observation = (
        lambda state: observation_preprocessor.transform(
            state.observation_by_agent[agent]
        )
    )
    actions, _, _ = robot_policy.compute_actions(
        np.stack(tuple(map(state_to_flat_observation, states))),
        explore=False
    )
    return dict(zip(states, map(Move.from_neural, actions), strict=True))






from __future__ import annotations

import os
from typing import Tuple, Optional, Dict, Mapping, Iterable, Any, Sequence
import pathlib
import inspect
import copy
import enum
import yaml
import datetime as datetime_module
import statistics
import pickle
import base64
import time
import abc
import lzma
import logging
import functools
import collections
import sys
import operator
import dataclasses
import functools
import itertools
import numbers
import pprint
import random
import re
import string
import contextlib

import more_itertools
import ray.rllib.env.multi_agent_env
import ray.rllib.algorithms.callbacks
from ray.rllib.utils.typing import PolicyID, MultiAgentDict, ModelWeights
from ray.rllib import algorithms
import ray.tune
import ray.rllib.algorithms.ppo
import gymnasium
import gymnasium.spaces
import numpy as np
import pysnooper
import click

from chicken_coop import county
from chicken_coop.county.thready_krueger import ThreadyKrueger
from chicken_coop.county import misc
from chicken_coop.county import trekking
from chicken_coop.county import csv_tools
from chicken_coop.county import policing
import chicken_coop
from chicken_coop.county.typing import Agent, RealNumber
from .zeitgeisting import Zeitgeist
from .stating import State
from .enving import Env
from . import defaults
from chicken_coop.coop_config import CoopConfig
from . import utils
from .callbackino import Callbackino

from .command_group import cli



def rollout_from_tune(config: dict[str, Any]) -> None:

    pure_config = {}
    coop_config_dict = {}
    config_arguments = set(inspect.signature(CoopConfig).parameters)
    for key, value in config.items():
        if key in config_arguments:
            coop_config_dict[key] = value
        else:
            pure_config[key] = value

    coop_config = CoopConfig(**coop_config_dict)
    return rollout(**pure_config, coop_config=coop_config,
                  trial_folder=pathlib.Path(ray.tune.get_trial_dir()),
                  report_to_tune=True, write_stdout_to_stdout=False)

def rollout(coop_config: CoopConfig, *,
            trial_folder: Optional[pathlib.Path] = None,
            report_to_tune: bool = False,
            write_stdout_to_stdout: bool = True,
            raw_visitor_policies: tuple[tuple[str, os.PathLike], ...] = (),
            argv: Optional[Iterable[str]] = None,
            ) -> None:

    trial_folder = trial_folder or pathlib.Path(ray.tune.get_trial_dir())

    samples_folder = trial_folder / 'samples'
    samples_folder.mkdir(parents=True, exist_ok=True)

    policy_snapshots_folder = trial_folder / 'policy_snapshots'
    policy_snapshots_folder.mkdir(parents=True, exist_ok=True)

    output_path: pathlib.Path = trial_folder / 'output.txt'
    reporters = [county.misc.JsonlRolloutReporter(trial_folder / 'rollout.jsonl')]

    if report_to_tune:
        reporters.append(county.misc.TuneRolloutReporter())

    if raw_visitor_policies:
        (visitor_mini_trek,) = {trekking.MiniTrek.get(policy_path)
                                for _, policy_path in raw_visitor_policies}
        visitor_policy_snapshot_by_policy_name = {
            policy_name: policing.PolicySnapshot.uncompickle_from_bytes(policy_path.read_bytes())
            for policy_name, policy_path in raw_visitor_policies
        }
        visitor_dominance_hierarchy = visitor_mini_trek.dominance_hierarchy
        assert visitor_dominance_hierarchy.is_complete

    else: # not raw_visitor_policies
        visitor_mini_trek = None
        visitor_dominance_hierarchy = \
                                   chicken_coop.DominanceHierarchy.get_trivial(coop_config.n_agents)
        visitor_policy_snapshot_by_policy_name = {}

    coop_config = dataclasses.replace(
        coop_config, visitor_dominance_hierarchy=visitor_dominance_hierarchy,
        i_visitor_agents=tuple(map(misc.get_int_from_mixed_string,
                                   visitor_policy_snapshot_by_policy_name)),
    )

    env = Env({'coop_config': coop_config})

    algorithm_config = (
        algorithms.ppo.PPOConfig()
        .training(
            train_batch_size=coop_config.train_batch_size,
            lr=coop_config.learning_rate,
        )
        .resources(
            num_gpus=0,
            num_gpus_per_worker=0,
        )
        .environment(
            env=Env,
            env_config=env.config,
            disable_env_checking=True,
        )
        .rollouts(
            create_env_on_local_worker=True,
            num_rollout_workers=coop_config.n_rollout_workers,
        )
        .reporting(
            keep_per_episode_custom_metrics=True,
        )
        .callbacks(
            Callbackino,
        )
        .multi_agent(
            policies={policy_name: ray.rllib.policy.policy.PolicySpec()
                      for policy_name in env.policies},
            policy_mapping_fn=coop_config.policy_mapping_fn,
            policies_to_train=tuple(
                policy_name for policy_name in env.policies if
                (not coop_config.freeze_visitors) or
                policy_name not in visitor_policy_snapshot_by_policy_name
            ),
        )
    )

    algorithm = algorithm_config.build()

    for visitor_policy_name, visitor_policy_snapshot in \
                                             visitor_policy_snapshot_by_policy_name.items():
        visitor_policy_snapshot.transplant_to_algorithm(algorithm, visitor_policy_name)


    if coop_config.flip_initial_weights:
        flipped_culture_snapshot = - policing.CultureSnapshot.import_from_algorithm(algorithm)
        flipped_culture_snapshot.transplant_to_algorithm(algorithm)

    # base_agentwise_measures = ('aggressiveness',)
    # simple_measure_fields = ()
    # simple_fields = ('generation', 'episode_reward', 'datetime')
    # agentwise_measure_fields = tuple(f'{agent}_{base_measure}' for agent in
                                     # self.coop_config.policy_by_agent for base_measure in
                                     # base_agentwise_measures)

    with (trial_folder / 'meta.yaml').open('w') as yaml_file:
        yaml.dump(
            {
                'coop_config': coop_config.get_nice_dict(for_yaml=True),
                'chicken_coop_repo_commit': county.misc.get_chicken_coop_repo_commit(),
                'chicken_coop_repo_is_clean': county.misc.get_chicken_coop_repo_is_clean(),
                'observation_fields': list(env.observation_space.keys()),
                'algorithm_class': type(algorithm).__name__,
                'raw_visitor_policies': {visitor_policy_name: str(visitor_policy_path) for
                                         visitor_policy_name, visitor_policy_path in
                                         raw_visitor_policies},
                'argv': sys.argv if argv is None else argv,
            },
            yaml_file,
            Dumper=misc.ListifyingDumper,
        )


    with contextlib.ExitStack() as exit_stack:
        if write_stdout_to_stdout:
            exit_stack.enter_context(county.misc.tee_stdout(output_path))
            stream = sys.stdout
        else:
            stream = exit_stack.enter_context(output_path.open('w'))

        zeitgeists = []

        for i_generation in range(coop_config.n_generations + 1):
            if i_generation == 0:
                # For fast failure in case there's a bug in `Callbackino`:
                Callbackino().on_train_result(algorithm=algorithm, result={})
            (samples_folder / f'{i_generation:04d}').write_text(
                Env.sample_episode_to_text(algorithm, coop_config=coop_config)
            )
            if (i_generation % coop_config.policy_snapshot_period == 0 or
                                                        i_generation == coop_config.n_generations):
                policy_snapshots_subfolder = (policy_snapshots_folder / f'{i_generation:04d}')
                policy_snapshots_subfolder.mkdir()
                culture_snapshot = policing.CultureSnapshot.import_from_algorithm(algorithm)
                for policy_name, policy_snapshot in culture_snapshot.items():
                    (policy_snapshots_subfolder / policy_name).write_bytes(
                                                               policy_snapshot.compickle_to_bytes())

            # generation_is_mature = i_generation >= 0.8 * n_generations
            if i_generation == 0:
                results = algorithm.evaluate()['evaluation']
            else:
                results = algorithm.train()
            from_array = lambda a_: Zeitgeist.from_array(a_, coop_config)
            zeitgeist = Zeitgeist.sum(
                map(from_array, results['custom_metrics']['zeitgeist_array'])
            )
            zeitgeists.append(zeitgeist)
            stream.write(f'Generation: {algorithm.iteration:04d}  '
                         f'Reward: {zeitgeist.population_portrait.reward:04.2f}\n')
            report = {
                'generation': algorithm.iteration,
                'datetime': datetime_module.datetime.now().isoformat(),
                'aggregated_dominance_hierarchy': (Zeitgeist.sum(zeitgeists[-5:]).
                                                           population_portrait.dominance_hierarchy),
                **zeitgeist.output,
            }
            for reporter in reporters:
                reporter.report(report)

    # Clean heavy files we don't need:
    for path in itertools.chain(trial_folder.glob('events.out.tfevents*'),
                                ((trial_folder / 'result.json'),)):
        path.unlink(missing_ok=True)


@cli.command()
@click.option('--moniker', type=str, default=None)
@click.option('-t', '--use-tune', is_flag=True, show_default=True)
@click.option('--n-tune-samples', type=int, default=defaults.DEFAULT_N_TUNE_SAMPLES,
              show_default=True)
@click.option('--visitor-policy', 'raw_visitor_policies',
              type=(str, click.Path(exists=True, dir_okay=False, resolve_path=True,
                                    path_type=pathlib.Path)),
              multiple=True, show_default=True)
@CoopConfig.add_options_to_click_command()
def run(*, moniker: Optional[str], use_tune: bool, n_tune_samples: int,
        raw_visitor_policies: tuple[tuple[str, str], ...], **raw_coop_config_kwargs) -> None:
    '''Run the Chicken Coop environment and train the agents.'''
    with trekking.Trek.create(extra_meta={'command': 'run', 'moniker': moniker}) as trek:
        county.init_ray()

        coop_config_kwargs = {}
        for key, values in raw_coop_config_kwargs.items():
            if not values:
                continue
            elif len(values) == 1:
                (value,) = values
                coop_config_kwargs[key] = value
            else:
                assert len(values) >= 2
                if not use_tune:
                    raise click.UsageError
                coop_config_kwargs[key] = ray.tune.grid_search(values)


        if use_tune:
            experiment = ray.tune.Experiment(
                name='chicken-coop',
                run=rollout_from_tune,
                config={
                    **coop_config_kwargs,
                    'raw_visitor_policies': raw_visitor_policies,
                    'argv': sys.argv,
                },
                num_samples=n_tune_samples,
                storage_path=str(tune_folder := trek.folder / 'tune_rollouts'),
                log_to_file=('stdout', 'stderr'),
                resources_per_trial=ray.tune.PlacementGroupFactory(
                    [{'CPU': 1.0}] + [{'CPU': 1.0}] * 2
                ),
            )
            experiment.dir_name = tune_folder

            analysis = ray.tune.run(experiment)
            df = misc.get_mean_dataframe_from_experiment_analysis(analysis)
            df.to_csv(trek.folder / 'tune_analysis.csv')

        else:
            rollout(coop_config=CoopConfig(**coop_config_kwargs),
                    raw_visitor_policies=raw_visitor_policies,
                    trial_folder=(trek.folder / 'rollout'))


@cli.command()
@click.option('--moniker', type=str, default=None)
@click.option('--visitor-trek', 'visitor_trek_path_string',
              type=click.Path(exists=True, dir_okay=True, file_okay=False, resolve_path=True,
                              path_type=pathlib.Path),
              show_default=True)
@click.option('--n-visitor-populations-per-n-visitor-agents', type=int,
              default=defaults.DEFAULT_N_VISITOR_POPULATIONS_PER_N_VISITOR_AGENTS,
              show_default=True)
@click.option('--n-transplants-per-visitor-population', type=int,
              default=defaults.DEFAULT_N_TRANSPLANTS_PER_VISITOR_POPULATION,
              show_default=True)
@CoopConfig.add_options_to_click_command(defaults={'learning_rate': (float(2e-05),),
                                                   'n_generations': (35,),
                                                   'train_batch_size': (256,),})
def transplant(*, moniker: Optional[str], visitor_trek_path_string: str,
               n_visitor_populations_per_n_visitor_agents: int,
               n_transplants_per_visitor_population: int, **raw_coop_config_kwargs) -> None:
    '''Transplant visitor agents into resident population.'''
    with trekking.Trek.create(extra_meta={'command': 'transplant', 'moniker': moniker}) as trek:
        county.init_ray()

        visitor_trek = trekking.Trek.get(visitor_trek_path_string, allow_single_trek=False)

        coop_config_kwargs = {}
        for key, values in raw_coop_config_kwargs.items():
            if not values:
                continue
            elif len(values) == 1:
                (value,) = values
                coop_config_kwargs[key] = value
            else:
                assert len(values) >= 2
                raise NotImplementedError("Grid search isn't implemented for `transplant`.")

        visitor_mini_treks = tuple(
            more_itertools.islice_extended(
                (visitor_mini_trek for visitor_mini_trek in visitor_trek.mini_treks
                 if visitor_mini_trek.dominance_hierarchy.is_complete),
                n_visitor_populations_per_n_visitor_agents
            )
        )
        assert len(visitor_mini_treks) == n_visitor_populations_per_n_visitor_agents
        fluff = []
        n_agents = CoopConfig(**coop_config_kwargs).n_agents
        for n_visitor_agents in range(n_agents - 1):
            for visitor_mini_trek in visitor_mini_treks:
                visitor_mini_trek: trekking.MiniTrek
                assert len(visitor_mini_trek.policy_path_by_name) == n_agents
                fluff.extend(
                    misc.get_wacky_choices(visitor_mini_trek.policy_path_by_name.items(),
                                           n=n_visitor_agents, k=n_transplants_per_visitor_population)
                )


        assert len(fluff) == ((n_agents - 1) * n_transplants_per_visitor_population *
                              n_visitor_populations_per_n_visitor_agents)
        assert set(collections.Counter(map(len, fluff)).values()) == \
                 {n_visitor_populations_per_n_visitor_agents * n_transplants_per_visitor_population}
        assert len(fluff[0]) == 0
        # Easing troubleshooting by making the first runs not be degenerate cases:
        for _ in range(100):
            if set(map(len, fluff[:4])) & {0, n_agents}:
                random.shuffle(fluff)
            else:
                break
        else:
            raise Exception
        raw_visitor_policies_grid_search = ray.tune.grid_search(fluff)


        experiment = ray.tune.Experiment(
            name='chicken-coop',
            run=rollout_from_tune,
            config={
                **coop_config_kwargs,
                'raw_visitor_policies': raw_visitor_policies_grid_search,
                'argv': sys.argv,
            },
            num_samples=1,
            storage_path=str(tune_folder := trek.folder / 'tune_rollouts'),
            log_to_file=('stdout', 'stderr'),
            resources_per_trial=ray.tune.PlacementGroupFactory(
                [{'CPU': 1.0}] + [{'CPU': 1.0}] * 2
            ),
        )
        experiment.dir_name = tune_folder

        analysis = ray.tune.run(experiment)

        df = misc.get_mean_dataframe_from_experiment_analysis(
            analysis,
        )

        df.to_csv(trek.folder / 'tune_analysis.csv')


